"""Core classes for the Destiny SDK, not exposed to package users."""

from typing import Self

from pydantic import BaseModel


class _JsonlFileInputMixIn(BaseModel):
    """
    A mixin class for models that are used at the top-level for entries in .jsonl files.

    This class is used to define a common interface for file input models.
    It is not intended to be used directly.
    """

    def to_jsonl(self) -> str:
        """
        Convert the model to a JSONL string.

        :return: The JSONL string representation of the model.
        :rtype: str
        """
        return self.model_dump_json(exclude_none=True)

    @classmethod
    def from_jsonl(cls, jsonl: str) -> Self:
        """
        Create an object from a JSONL string.

        :param jsonl: The JSONL string to parse.
        :type jsonl: str
        :return: The created object.
        :rtype: Self
        """
        return cls.model_validate_json(jsonl)



"""Schemas that define inputs/outputs for robots."""

from enum import StrEnum, auto
from typing import Annotated, Any, Self

from pydantic import UUID4, BaseModel, ConfigDict, Field, HttpUrl, model_validator

from destiny_sdk.core import _JsonlFileInputMixIn
from destiny_sdk.enhancements import Enhancement
from destiny_sdk.references import Reference


class RobotError(BaseModel):
    """A record of something going wrong with the robot."""

    message: Annotated[
        str,
        Field(
            description=(
                "Message which describes the error encountered during processing"
            )
        ),
    ]


class LinkedRobotError(_JsonlFileInputMixIn, RobotError):
    """
    A record of something going wrong when processing an individual reference.

    Used in results for batch requests - in single requests, the reference
    id is derived from the request id.
    """

    reference_id: UUID4 = Field(
        description="The ID of the reference which caused the error."
    )


class RobotResult(BaseModel):
    """The result of a robot request which is returned to the repo."""

    request_id: UUID4
    error: RobotError | None = Field(
        default=None,
        description="Error the robot encountered while creating enhancement.",
    )
    enhancement: Enhancement | None = Field(
        default=None, description="An enhancement to create"
    )

    @model_validator(mode="after")
    def validate_error_or_enhancement_set(self) -> Self:
        """Validate that a robot result has either an error or an enhancement set."""
        if (self.error is None) == (self.enhancement is None):
            msg = """
            exactly one of 'error' or 'enhancement' must be provided
            """
            raise ValueError(msg)
        return self


class BatchRobotResult(BaseModel):
    """Used to indicate to the repository that the robot has finished processing."""

    request_id: UUID4
    error: RobotError | None = Field(
        default=None,
        description="""
Error the robot encountered while creating enhancements. If this field is populated,
we assume the entire batch or request failed, rather than an individual reference.
If there was an error with processing an individual reference, it should be passed in
the result file and this field should be left as None. Vice-versa, if this field is
None, the repository will assume that the result file is ready for processing.
""",
    )


class BatchRobotResultValidationEntry(_JsonlFileInputMixIn, BaseModel):
    """A single entry in the validation result file for a batch enhancement request."""

    reference_id: UUID4 | None = Field(
        default=None,
        description=(
            "The ID of the reference which was enhanced. "
            "If this is empty, the BatchEnhancementResultEntry could not be parsed."
        ),
    )
    error: str | None = Field(
        default=None,
        description=(
            "Error encountered during the enhancement process for this reference. "
            "If this is empty, the enhancement was successfully created."
        ),
    )


class RobotRequest(BaseModel):
    """An enhancement request from the repo to a robot."""

    id: UUID4
    reference: Reference = Field(
        description=(
            "Reference to be enhanced, includes identifiers and existing enhancments."
        )
    )
    extra_fields: dict | None = Field(
        default=None,
        description="Extra fields to pass to the robot. TBC.",
    )


#: The result for a single reference when processed by a batch enhancement request.
#: This is a single entry in the result file.
BatchEnhancementResultEntry = Annotated[
    Enhancement | LinkedRobotError,
    Field(),
]


class BatchRobotRequest(BaseModel):
    """A batch enhancement request from the repo to a robot."""

    id: UUID4
    reference_storage_url: HttpUrl = Field(
        description="""
The URL at which the set of references are stored. The file is a jsonl
with each line formatted according to
:class:`Reference <libs.sdk.src.destiny_sdk.references.Reference>`, one
reference per line.
Each reference may have identifiers or enhancements attached, as
required by the robot.
If the URL expires, a new one can be generated using
``GET /references/enhancement/batch/<batch_request_id>``.
"""
    )
    result_storage_url: HttpUrl = Field(
        description="""
The URL at which the set of enhancements are to be stored. The file is to be a jsonl
with each line formatted according to
:class:`BatchEnhancementResultEntry <libs.sdk.src.destiny_sdk.robots.BatchEnhancementResultEntry>`.
If the URL expires, a new one can be generated using
``GET /references/enhancement/batch/<batch_request_id>``.
"""  # noqa: E501
    )
    extra_fields: dict | None = Field(
        default=None,
        description="Extra fields to pass to the robot. TBC.",
    )


class EnhancementRequestStatus(StrEnum):
    """
    The status of an enhancement request.

    **Allowed values**:
    - `received`: Enhancement request has been received.
    - `accepted`: Enhancement request has been accepted.
    - `rejected`: Enhancement request has been rejected.
    - `failed`: Enhancement failed to create.
    - `completed`: Enhancement has been created.
    """

    RECEIVED = auto()
    ACCEPTED = auto()
    REJECTED = auto()
    FAILED = auto()
    COMPLETED = auto()


class _EnhancementRequestBase(BaseModel):
    """
    Base enhancement request class.

    An enhancement request is a request to create an enhancement on a reference.
    It contains the reference and the robot to be used to create the enhancement.
    """

    reference_id: UUID4 = Field(description="The ID of the reference to be enhanced.")
    robot_id: UUID4 = Field(
        description="The robot to be used to create the enhancement."
    )
    source: str | None = Field(
        default=None,
        description="The source of the batch enhancement request.",
    )

    enhancement_parameters: dict | None = Field(
        default=None, description="Information needed to create the enhancement. TBC."
    )


class EnhancementRequestIn(_EnhancementRequestBase):
    """The model for requesting an enhancement on specific reference."""


class EnhancementRequestRead(_EnhancementRequestBase):
    """Core enhancement request class."""

    id: UUID4
    request_status: EnhancementRequestStatus = Field(
        description="The status of the request to create an enhancement",
    )
    error: str | None = Field(
        default=None,
        description="Error encountered during the enhancement process",
    )


class BatchEnhancementRequestStatus(StrEnum):
    """
    The status of an enhancement request.

    **Allowed values**:
    - `received`: Enhancement request has been received by the repo.
    - `accepted`: Enhancement request has been accepted by the robot.
    - `rejected`: Enhancement request has been rejected by the robot.
    - `partial_failed`: Some enhancements failed to create.
    - `failed`: All enhancements failed to create.
    - `importing`: Enhancements have been received by the repo and are being imported.
    - `indexing`: Enhancements have been imported and are being indexed.
    - `indexing_failed`: Enhancements have been imported but indexing failed.
    - `completed`: All enhancements have been created.
    """

    RECEIVED = auto()
    ACCEPTED = auto()
    REJECTED = auto()
    PARTIAL_FAILED = auto()
    FAILED = auto()
    IMPORTING = auto()
    INDEXING = auto()
    INDEXING_FAILED = auto()
    COMPLETED = auto()


class _BatchEnhancementRequestBase(BaseModel):
    """
    Base batch enhancement request class.

    A batch enhancement request is a request to create multiple enhancements.
    """

    robot_id: UUID4 = Field(
        description="The robot to be used to create the enhancements."
    )
    reference_ids: list[UUID4] = Field(
        description="The IDs of the references to be enhanced."
    )
    source: str | None = Field(
        default=None,
        description="The source of the batch enhancement request.",
    )


class BatchEnhancementRequestIn(_BatchEnhancementRequestBase):
    """The model for requesting multiple enhancements on specific references."""


class BatchEnhancementRequestRead(_BatchEnhancementRequestBase):
    """Core batch enhancement request class."""

    id: UUID4
    request_status: BatchEnhancementRequestStatus = Field(
        description="The status of the request to create enhancements",
    )
    reference_data_url: HttpUrl | None = Field(
        default=None,
        description="""
The URL at which the set of references are stored. The file is a jsonl with each line
formatted according to
:class:`Reference <libs.sdk.src.destiny_sdk.references.Reference>`.
, one reference per line.
Each reference may have identifiers or enhancements attached, as
required by the robot.
If the URL expires, a new one can be generated using
``GET /references/enhancement/batch/<batch_request_id>``.
        """,
    )
    result_storage_url: HttpUrl | None = Field(
        default=None,
        description="""
The URL at which the set of enhancements are stored. The file is to be a jsonl
with each line formatted according to
:class:`BatchEnhancementResultEntry <libs.sdk.src.destiny_sdk.robots.BatchEnhancementResultEntry>`.
This field is only relevant to robots.
If the URL expires, a new one can be generated using
``GET /references/enhancement/batch/<batch_request_id>``.
        """,  # noqa: E501
    )
    validation_result_url: HttpUrl | None = Field(
        default=None,
        description="""
The URL at which the result of the batch enhancement request is stored.
This file is a txt file, one line per reference, with either an error
or a success message.
If the URL expires, a new one can be generated using
``GET /references/enhancement/batch/<batch_request_id>``.
        """,
    )
    error: str | None = Field(
        default=None,
        description="Error encountered during the enhancement process. This "
        "is only used if the entire batch enhancement request failed, rather than an "
        "individual reference. If there was an error with processing an individual "
        "reference, it is passed in the validation result file.",
    )


class _RobotBase(BaseModel):
    """
    Base Robot class.

    A Robot is a provider of enhancements to destiny repository
    """

    model_config = ConfigDict(extra="forbid")  # Forbid extra fields on robot models

    name: str = Field(description="The name of the robot, must be unique.")
    base_url: HttpUrl = Field(
        description="The base url of the robot. The robot must implement endpoints "
        "base_url/single for the enhancement of single references and "
        "base_url/batch for batch enhancements of references.",
    )
    description: str = Field(
        description="Description of the enhancement the robot provides."
    )
    owner: str = Field(description="The owner/publisher of the robot.")


class RobotIn(_RobotBase):
    """The model for registering a new robot."""


class Robot(_RobotBase):
    """Then model for a registered robot."""

    id: UUID4 = Field(
        description="The id of the robot provided by destiny repository. "
        "Used as the client_id when sending HMAC authenticated requests."
    )


class ProvisionedRobot(Robot):
    """
    The model for a provisioned robot.

    Used only when a robot is initially created,
    or when cycling a robot's client_secret.
    """

    client_secret: str = Field(
        description="The client secret of the robot, used as the secret key "
        "when sending HMAC authenticated requests."
    )


class _RobotAutomationBase(BaseModel):
    """Base Robot Automation class."""

    robot_id: UUID4 = Field(
        description="The ID of the robot that will be used to enhance the reference."
    )
    query: dict[str, Any] = Field(
        description="The percolator query that will be used to match references "
        " or enhancements against."
    )


class RobotAutomationIn(_RobotAutomationBase):
    """
    Automation model for a robot.

    This is used as a source of truth for an Elasticsearch index that percolates
    references or enhancements against the queries. If a query matches, a request
    is sent to the specified robot to perform the enhancement.
    """


class RobotAutomation(_RobotAutomationBase):
    """
    Core Robot Automation class.

    This is used as a source of truth for an Elasticsearch index that percolates
    references or enhancements against the queries. If a query matches, a request
    is sent to the specified robot to perform the enhancement.
    """

    id: UUID4 = Field(
        description="The ID of the robot automation.",
    )



"""Identifier classes for the Destiny SDK."""

from enum import StrEnum, auto
from typing import Annotated, Literal

from pydantic import UUID4, BaseModel, Field, field_validator


class ExternalIdentifierType(StrEnum):
    """
    The type of identifier used to identify a reference.

    This is used to identify the type of identifier used in the `ExternalIdentifier`
    class.
    **Allowed values**:
    - `doi`: A DOI (Digital Object Identifier) which is a unique identifier for a
    document.
    - `pmid`: A PubMed ID which is a unique identifier for a document in PubMed.
    - `openalex`: An OpenAlex ID which is a unique identifier for a document in
    OpenAlex.
    - `other`: Any other identifier not defined. This should be used sparingly.
    """

    DOI = auto()
    PM_ID = auto()
    OPEN_ALEX = auto()
    OTHER = auto()


class DOIIdentifier(BaseModel):
    """An external identifier representing a DOI."""

    identifier: str = Field(
        description="The DOI of the reference.",
        pattern=r"^10\.\d{4,9}/[-._;()/:a-zA-Z0-9%<>\[\]+&]+$",
    )
    identifier_type: Literal[ExternalIdentifierType.DOI] = Field(
        ExternalIdentifierType.DOI, description="The type of identifier used."
    )

    @field_validator("identifier", mode="before")
    @classmethod
    def remove_doi_url(cls, value: str) -> str:
        """Remove the URL part of the DOI if it exists."""
        return (
            value.removeprefix("http://doi.org/")
            .removeprefix("https://doi.org/")
            .strip()
        )


class PubMedIdentifier(BaseModel):
    """An external identifier representing a PubMed ID."""

    identifier: int = Field(description="The PubMed ID of the reference.")
    identifier_type: Literal[ExternalIdentifierType.PM_ID] = Field(
        ExternalIdentifierType.PM_ID, description="The type of identifier used."
    )


class OpenAlexIdentifier(BaseModel):
    """An external identifier representing an OpenAlex ID."""

    identifier: str = Field(
        description="The OpenAlex ID of the reference.", pattern=r"^W\d+$"
    )
    identifier_type: Literal[ExternalIdentifierType.OPEN_ALEX] = Field(
        ExternalIdentifierType.OPEN_ALEX, description="The type of identifier used."
    )

    @field_validator("identifier", mode="before")
    @classmethod
    def remove_open_alex_url(cls, value: str) -> str:
        """Remove the OpenAlex URL if it exists."""
        return (
            value.removeprefix("http://openalex.org/")
            .removeprefix("https://openalex.org/")
            .strip()
        )


class OtherIdentifier(BaseModel):
    """An external identifier not otherwise defined by the repository."""

    identifier: str = Field(description="The identifier of the reference.")
    identifier_type: Literal[ExternalIdentifierType.OTHER] = Field(
        ExternalIdentifierType.OTHER, description="The type of identifier used."
    )
    other_identifier_name: str = Field(
        description="The name of the undocumented identifier type."
    )


#: Union type for all external identifiers.
ExternalIdentifier = Annotated[
    DOIIdentifier | PubMedIdentifier | OpenAlexIdentifier | OtherIdentifier,
    Field(discriminator="identifier_type"),
]


class LinkedExternalIdentifier(BaseModel):
    """An external identifier which identifies a reference."""

    identifier: ExternalIdentifier = Field(
        description="The identifier of the reference.",
        discriminator="identifier_type",
    )
    reference_id: UUID4 = Field(
        description="The ID of the reference this identifier identifies."
    )



import uuid
from datetime import date

import destiny_sdk
import pytest
from pydantic import ValidationError


def test_bibliographic_metadata_enhancement_valid():
    # Create valid bibliographic content
    bibliographic = destiny_sdk.enhancements.BibliographicMetadataEnhancement(
        enhancement_type=destiny_sdk.enhancements.EnhancementType.BIBLIOGRAPHIC,
        authorship=[],
        cited_by_count=10,
        created_date=date(2020, 1, 1),
        publication_date=date(2050, 1, 2),
        publication_year=2020,
        publisher="Test Publisher",
        title="Test Title",
    )
    enhancement = destiny_sdk.enhancements.Enhancement(
        id=uuid.uuid4(),
        source="test_source",
        visibility="public",
        robot_version="1.0",
        enhancement_type=destiny_sdk.enhancements.EnhancementType.BIBLIOGRAPHIC,
        content=bibliographic,
        reference_id=uuid.uuid4(),
    )
    assert (
        enhancement.content.enhancement_type
        == destiny_sdk.enhancements.EnhancementType.BIBLIOGRAPHIC
    )


def test_abstract_content_enhancement_valid():
    # Create valid abstract content
    abstract_content = destiny_sdk.enhancements.AbstractContentEnhancement(
        enhancement_type=destiny_sdk.enhancements.EnhancementType.ABSTRACT,
        process=destiny_sdk.enhancements.AbstractProcessType.UNINVERTED,
        abstract="This is a test abstract.",
    )
    enhancement = destiny_sdk.enhancements.Enhancement(
        id=uuid.uuid4(),
        source="test_source",
        visibility="public",
        robot_version="2.0",
        enhancement_type=destiny_sdk.enhancements.EnhancementType.ABSTRACT,
        content=abstract_content,
        reference_id=uuid.uuid4(),
    )
    assert enhancement.content.abstract == "This is a test abstract."


def test_annotation_enhancement_valid():
    # Create valid annotation content
    annotation1 = destiny_sdk.enhancements.BooleanAnnotation(
        annotation_type=destiny_sdk.enhancements.AnnotationType.BOOLEAN,
        scheme="openalex:topic",
        value=True,
        label="Machine Learning",
        score=0.95,
        data={"confidence": 0.95},
    )
    annotations_content = destiny_sdk.enhancements.AnnotationEnhancement(
        enhancement_type=destiny_sdk.enhancements.EnhancementType.ANNOTATION,
        annotations=[annotation1],
    )
    enhancement = destiny_sdk.enhancements.Enhancement(
        id=uuid.uuid4(),
        source="test_source",
        visibility="public",
        robot_version="1.5",
        enhancement_type=destiny_sdk.enhancements.EnhancementType.ANNOTATION,
        content=annotations_content,
        reference_id=uuid.uuid4(),
    )
    assert enhancement.content.annotations[0].label == "Machine Learning"


def test_location_enhancement_valid():
    # Create valid location content
    location = destiny_sdk.enhancements.Location(
        is_oa=True,
        version="publishedVersion",
        landing_page_url="https://example.com",
        pdf_url="https://example.com/doc.pdf",
        license="cc-by",
        extra={"note": "Accessible"},
    )
    location_content = destiny_sdk.enhancements.LocationEnhancement(
        enhancement_type=destiny_sdk.enhancements.EnhancementType.LOCATION,
        locations=[location],
    )
    enhancement = destiny_sdk.enhancements.Enhancement(
        id=uuid.uuid4(),
        source="test_source",
        visibility="public",
        robot_version="1.2",
        enhancement_type=destiny_sdk.enhancements.EnhancementType.LOCATION,
        content=location_content,
        reference_id=uuid.uuid4(),
    )
    assert enhancement.content.locations[0].license == "cc-by"


def test_empty_annotation_enhancement_errors():
    # Test that an empty annotations list raises a validation error
    with pytest.raises(
        ValidationError, match="List should have at least 1 item after validation"
    ):
        destiny_sdk.enhancements.AnnotationEnhancement(
            enhancement_type=destiny_sdk.enhancements.EnhancementType.ANNOTATION,
            annotations=[],
        )


def test_empty_location_enhancement_errors():
    # Test that an empty locations list raises a validation error
    with pytest.raises(
        ValidationError, match="List should have at least 1 item after validation"
    ):
        destiny_sdk.enhancements.LocationEnhancement(
            enhancement_type=destiny_sdk.enhancements.EnhancementType.LOCATION,
            locations=[],
        )



import uuid

import destiny_sdk
import pytest
from pydantic import HttpUrl, ValidationError


def test_enhancement_request_valid():
    enhancement_request = destiny_sdk.robots.EnhancementRequestRead(
        id=uuid.uuid4(),
        reference_id=uuid.uuid4(),
        reference=destiny_sdk.references.Reference(
            id=uuid.uuid4(), visibility=destiny_sdk.visibility.Visibility.RESTRICTED
        ),
        robot_id=uuid.uuid4(),
        request_status=destiny_sdk.robots.EnhancementRequestStatus.RECEIVED,
    )

    assert (
        enhancement_request.request_status
        == destiny_sdk.robots.EnhancementRequestStatus.RECEIVED
    )
    assert enhancement_request.enhancement_parameters is None
    assert enhancement_request.error is None


def test_provisioned_robot_valid():
    provisioned_robot = destiny_sdk.robots.ProvisionedRobot(
        id=uuid.uuid4(),
        base_url=HttpUrl("https://www.domo-arigato-mr-robo.to"),
        name="Mr. Roboto",
        description="I have come to help you with your problems",
        owner="Styx",
        client_secret="secret, secret, I've got a secret",
    )

    assert provisioned_robot.owner == "Styx"


def test_robot_models_reject_any_extra_fields():
    with pytest.raises(ValidationError):
        destiny_sdk.robots.Robot(
            base_url=HttpUrl("https://www.domo-arigato-mr-robo.to"),
            name="Mr. Roboto",
            description="I have come to help you with your problems",
            owner="Styx",
            client_secret="I'm not allowed in this model",
        )


def test_robot_request_does_not_require_extra_fields():
    reference = destiny_sdk.references.Reference(
        id=uuid.uuid4(), visibility=destiny_sdk.visibility.Visibility.RESTRICTED
    )

    # Don't pass any extra fields
    robot_request = destiny_sdk.robots.RobotRequest(
        id=uuid.uuid4(), reference=reference
    )

    assert not robot_request.extra_fields



"""Import process classes for the Destiny SDK."""

import datetime
from enum import StrEnum, auto

from pydantic import (
    UUID4,
    BaseModel,
    Field,
    HttpUrl,
    PastDatetime,
)


class ImportRecordStatus(StrEnum):
    """
    Describes the status of an import record.

    - `created`: Created, but no processing has started.
    - `started`: Processing has started on the batch.
    - `completed`: Processing has been completed.
    - `cancelled`: Processing was cancelled by calling the API.
    """

    CREATED = auto()
    STARTED = auto()
    COMPLETED = auto()
    CANCELLED = auto()


class ImportBatchStatus(StrEnum):
    """
    Describes the status of an import batch.

    - `created`: Created, but no processing has started.
    - `started`: Processing has started on the batch.
    - `failed`: Processing has failed.
    - `retrying`: Processing has failed, but is being retried.
    - `indexing`: The imports have been saved and are being indexed.
    - `indexing_failed`: The imports have been saved but were not indexed.
    - `completed`: Processing has been completed.
    - `cancelled`: Processing was cancelled by calling the API.
    """

    CREATED = auto()
    STARTED = auto()
    RETRYING = auto()
    FAILED = auto()
    INDEXING = auto()
    INDEXING_FAILED = auto()
    COMPLETED = auto()
    CANCELLED = auto()


class CollisionStrategy(StrEnum):
    """
    The strategy to use when an identifier collision is detected.

    Identifier collisions are detected on ``identifier_type`` and ``identifier``
    (and ``other_identifier_name`` where relevant) already present in the database.

    Enhancement collisions are detected on an entry with matching ``enhancement_type``
    and ``source`` already being present on the collided reference.

    - `discard`: Do nothing with the incoming reference.
    - `fail`: Do nothing with the incoming reference and mark it as failed. This
      allows the importing process to "follow up" on the failure.
    - `merge_aggressive`: Prioritize the incoming reference's identifiers and
      enhancements in the merge.
    - `merge_defensive`: Prioritize the existing reference's identifiers and
      enhancements in the merge.
    - `append`: Performs an aggressive merge of identifiers, and an append of
      enhancements.
    - `overwrite`: Performs an aggressive merge of identifiers, and an overwrite of
      enhancements (deleting existing and recreating what is imported). This should
      be used sparingly and carefully.
    """

    DISCARD = auto()
    FAIL = auto()
    MERGE_AGGRESSIVE = auto()
    MERGE_DEFENSIVE = auto()
    APPEND = auto()
    OVERWRITE = auto()


class ImportResultStatus(StrEnum):
    """
    Describes the status of an import result.

    - `created`: Created, but no processing has started.
    - `started`: The reference is currently being processed.
    - `completed`: The reference has been created.
    - `partially_failed`: The reference was created but one or more enhancements or
      identifiers failed to be added. See the result's `failure_details` field for
      more information.
    - `failed`: The reference failed to be created. See the result's `failure_details`
      field for more information.
    - `cancelled`: Processing was cancelled by calling the API.
    """

    CREATED = auto()
    STARTED = auto()
    COMPLETED = auto()
    CANCELLED = auto()
    PARTIALLY_FAILED = auto()
    FAILED = auto()


class _ImportRecordBase(BaseModel):
    """Base import record class."""

    search_string: str | None = Field(
        default=None,
        description="The search string used to produce this import",
    )
    searched_at: PastDatetime = Field(
        default_factory=lambda: datetime.datetime.now(tz=datetime.UTC),
        description="""
The timestamp (including timezone) at which the search which produced
this import was conducted. If no timezone is included, the timestamp
is assumed to be in UTC.
        """,
    )
    processor_name: str = Field(
        description="The name of the processor that is importing the data."
    )
    processor_version: str = Field(
        description="The version of the processor that is importing the data."
    )
    notes: str | None = Field(
        default=None,
        description="""
Any additional notes regarding the import (eg. reason for importing, known
issues).
        """,
    )
    expected_reference_count: int = Field(
        description="""
The number of references expected to be included in this import.
-1 is accepted if the number is unknown.
""",
        ge=-1,
    )
    source_name: str = Field(
        description="The source of the reference being imported (eg. Open Alex)"
    )


class ImportRecordIn(_ImportRecordBase):
    """Input for creating an import record."""


class ImportRecordRead(_ImportRecordBase):
    """Core import record class."""

    id: UUID4 = Field(
        description="The ID of the import record",
    )
    status: ImportRecordStatus = Field(
        ImportRecordStatus.CREATED,
        description="The status of the import record",
    )
    batches: list["ImportBatchRead"] | None = Field(
        default=None,
        description="A list of batches for the import record",
    )


class _ImportBatchBase(BaseModel):
    """The base class for import batches."""

    collision_strategy: CollisionStrategy = Field(
        default=CollisionStrategy.FAIL,
        description="""
The strategy to use for each reference when an identifier collision occurs.
Default is `fail`, which allows the importing process to "follow up" on the collision.
        """,
    )
    storage_url: HttpUrl = Field(
        description="""
The URL at which the set of references for this batch are stored. The file is a jsonl
with each line formatted according to
:class:`ReferenceFileInput <libs.sdk.src.destiny_sdk.references.ReferenceFileInput>`.
    """,
    )
    callback_url: HttpUrl | None = Field(
        default=None,
        description="""
The URL to which the processor should send a callback when the batch has been processed.
        """,
    )


class ImportBatchIn(_ImportBatchBase):
    """Input for creating an import batch."""


class ImportBatchRead(_ImportBatchBase):
    """Core import batch class."""

    id: UUID4 = Field(
        description="The ID of the import batch",
    )
    status: ImportBatchStatus = Field(
        default=ImportBatchStatus.CREATED, description="The status of the batch."
    )
    import_record_id: UUID4 = Field(
        description="The ID of the import record this batch is associated with"
    )
    import_record: ImportRecordRead | None = Field(
        default=None, description="The parent import record."
    )
    import_results: list["ImportResultRead"] | None = Field(
        default=None, description="The results from processing the batch."
    )


class ImportBatchSummary(_ImportBatchBase):
    """A view for an import batch that includes a summary of its results."""

    id: UUID4 = Field(
        description="""
The identifier of the batch.
""",
    )

    import_batch_id: UUID4 = Field(description="The ID of the batch being summarised")

    import_batch_status: ImportBatchStatus = Field(
        description="The status of the batch being summarised"
    )

    results: dict[ImportResultStatus, int] = Field(
        description="A count of references by their current import status."
    )
    failure_details: list[str] | None = Field(
        description="""
        The details of the failures that occurred.
        Each failure will start with `"Entry x"` where x is the line number of the
        jsonl object attempted to be imported.
        """,
    )


class ImportResultRead(BaseModel):
    """Core import result class."""

    id: UUID4 = Field(description="The ID of the import result.")
    reference_id: UUID4 | None = Field(
        default=None,
        description="The ID of the reference created by this import result.",
    )
    failure_details: str | None = Field(
        default=None,
        description="The details of the failure, if the import result failed.",
    )
    import_batch: ImportBatchRead | None = Field(
        default=None, description="The parent import batch."
    )



import asyncio
import logging
from typing import Any

import pytest

MIGRATION_TASK: asyncio.Task | None = None

logging.getLogger("asyncio").setLevel("DEBUG")


@pytest.fixture(scope="session", autouse=True)
def anyio_backend() -> tuple[str, dict[str, Any]]:
    """Specify the anyio backend for async tests."""
    return "asyncio", {"use_uvloop": True}



"""Reference classes for the Destiny SDK."""

from typing import Self

from pydantic import UUID4, BaseModel, Field, TypeAdapter

from destiny_sdk.core import _JsonlFileInputMixIn
from destiny_sdk.enhancements import Enhancement, EnhancementFileInput
from destiny_sdk.identifiers import ExternalIdentifier
from destiny_sdk.visibility import Visibility

external_identifier_adapter = TypeAdapter(ExternalIdentifier)


class Reference(_JsonlFileInputMixIn, BaseModel):
    """Core reference class."""

    visibility: Visibility = Field(
        default=Visibility.PUBLIC,
        description="The level of visibility of the reference",
    )
    id: UUID4 = Field(
        description="The ID of the reference",
    )
    identifiers: list[ExternalIdentifier] | None = Field(
        default=None,
        description="A list of `ExternalIdentifiers` for the Reference",
    )
    enhancements: list[Enhancement] | None = Field(
        default=None,
        description="A list of enhancements for the reference",
    )

    @classmethod
    def from_es(cls, es_reference: dict) -> Self:
        """Create a Reference from an Elasticsearch document."""
        return cls(
            id=es_reference["_id"],
            visibility=Visibility(es_reference["_source"]["visibility"]),
            identifiers=[
                external_identifier_adapter.validate_python(identifier)
                for identifier in es_reference["_source"].get("identifiers", [])
            ],
            enhancements=[
                Enhancement.model_validate(
                    enhancement | {"reference_id": es_reference["_id"]},
                )
                for enhancement in es_reference["_source"].get("enhancements", [])
            ],
        )


class ReferenceFileInput(_JsonlFileInputMixIn, BaseModel):
    """Enhancement model used to marshall a file input."""

    visibility: Visibility = Field(
        default=Visibility.PUBLIC,
        description="The level of visibility of the reference",
    )
    identifiers: list[ExternalIdentifier] | None = Field(
        default=None,
        description="A list of `ExternalIdentifiers` for the Reference",
    )
    enhancements: list[EnhancementFileInput] | None = Field(
        default=None,
        description="A list of enhancements for the reference",
    )



"""Tests for HMAC Authentication."""

import time
import uuid
from collections.abc import AsyncGenerator

import destiny_sdk
import pytest
from destiny_sdk.client import create_signature
from fastapi import APIRouter, Depends, FastAPI, status
from httpx import ASGITransport, AsyncClient

TEST_SECRET_KEY = "dlfskdfhgk8ei346oiehslkdfrerikfglser934utofs"
TEST_CLIENT_ID = uuid.uuid4()
REQUEST_BODY = b'{"message": "info"}'


@pytest.fixture
def hmac_app() -> FastAPI:
    """
    Create a FastAPI application instance for testing HMAC authentication.

    Returns:
        FastAPI: FastAPI app with test router configured with HMAC auth.

    """
    app = FastAPI(title="Test HMAC Auth")
    auth = destiny_sdk.auth.HMACAuth(secret_key=TEST_SECRET_KEY)

    def __endpoint() -> dict:
        return {"message": "ok"}

    router = APIRouter(prefix="/test", dependencies=[Depends(auth)])
    router.add_api_route(
        path="/hmac/",
        methods=["POST"],
        status_code=status.HTTP_200_OK,
        endpoint=__endpoint,
    )

    app.include_router(router)
    return app


@pytest.fixture
async def client(hmac_app: FastAPI) -> AsyncGenerator[AsyncClient]:
    """
    Create a test client for the FastAPI application.

    Args:
        app (FastAPI): FastAPI application instance.

    Returns:
        TestClient: Test client for the FastAPI application.

    """
    async with AsyncClient(
        transport=ASGITransport(app=hmac_app),
        base_url="http://test",
    ) as client:
        yield client


async def test_hmac_authentication_happy_path(client: AsyncClient):
    """Test authentication is successful when signature is correct."""
    auth = destiny_sdk.client.HMACSigningAuth(
        secret_key=TEST_SECRET_KEY, client_id=TEST_CLIENT_ID
    )

    response = await client.post("test/hmac/", content=REQUEST_BODY, auth=auth)

    assert response.status_code == status.HTTP_200_OK


async def test_hmac_authentication_no_signature(client: AsyncClient):
    """Test authentication fails if the signature is not included."""
    response = await client.post("test/hmac/", content=REQUEST_BODY)

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "Authorization header missing" in response.json()["detail"]


async def test_hmac_authentication_wrong_auth_type(client: AsyncClient):
    """Test authentication fails if the signature is not included."""
    response = await client.post(
        "test/hmac/",
        content=REQUEST_BODY,
        headers={"Authorization": "Bearer nonsense-token"},
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "type not supported" in response.json()["detail"]


async def test_hmac_authentication_no_client_id(client: AsyncClient):
    """Test authentication fails when no client id is provided"""
    signature = create_signature(
        secret_key=TEST_SECRET_KEY,
        request_body=REQUEST_BODY,
        client_id=TEST_CLIENT_ID,
        timestamp=time.time(),
    )

    response = await client.post(
        "test/hmac/",
        headers={"Authorization": f"Signature {signature}"},
        content=REQUEST_BODY,
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "X-Client-Id header missing" in response.json()["detail"]


async def test_hmac_authentication_invalid_client_id_format(client: AsyncClient):
    """Test authentication fails when no client id is provided"""
    signature = create_signature(
        secret_key=TEST_SECRET_KEY,
        request_body=REQUEST_BODY,
        client_id=TEST_CLIENT_ID,
        timestamp=time.time(),
    )

    response = await client.post(
        "test/hmac/",
        headers={
            "Authorization": f"Signature {signature}",
            "X-Client-Id": "not-a-uuid",
        },
        content=REQUEST_BODY,
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "Invalid format for client id" in response.json()["detail"]


async def test_hmac_authentication_no_timestamp(client: AsyncClient):
    """Test authentication fails when no client id is provided"""
    signature = create_signature(
        secret_key=TEST_SECRET_KEY,
        request_body=REQUEST_BODY,
        client_id=TEST_CLIENT_ID,
        timestamp=time.time(),
    )

    response = await client.post(
        "test/hmac/",
        headers={
            "Authorization": f"Signature {signature}",
            "X-Client-Id": f"{TEST_CLIENT_ID}",
        },
        content=REQUEST_BODY,
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "X-Request-Timestamp header missing" in response.json()["detail"]


async def test_hmac_authentication_request_too_old(client: AsyncClient):
    """Test authentication fails when no client id is provided"""
    six_minutes = 60 * 6
    signature = create_signature(
        secret_key=TEST_SECRET_KEY,
        request_body=REQUEST_BODY,
        client_id=TEST_CLIENT_ID,
        timestamp=time.time(),
    )

    response = await client.post(
        "test/hmac/",
        headers={
            "Authorization": f"Signature {signature}",
            "X-Client-Id": f"{TEST_CLIENT_ID}",
            "X-Request-Timestamp": f"{time.time() - six_minutes}",
        },
        content=REQUEST_BODY,
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "Request timestamp has expired." in response.json()["detail"]


async def test_hmac_authentication_incorrect_signature(client: AsyncClient):
    """Test authentication fails when the signature does not match."""
    response = await client.post(
        "test/hmac/",
        headers={
            "Authorization": "Signature nonsense-signature",
            "X-Client-Id": f"{TEST_CLIENT_ID}",
            "X-Request-Timestamp": f"{time.time()}",
        },
        content=REQUEST_BODY,
    )

    assert response.status_code == status.HTTP_401_UNAUTHORIZED
    assert "Signature" in response.json()["detail"]



import uuid

import destiny_sdk


def test_es_parsing():
    """Test that the Reference model can be parsed from an Elasticsearch document."""
    es_reference = {
        "_index": "destiny-repository-staging-reference",
        "_id": "7c54f72e-8833-484f-9000-4f403b13a243",
        "_score": 1.7860229,
        "_ignored": ["enhancements.content.abstract.keyword"],
        "_source": {
            "visibility": "public",
            "identifiers": [
                {"identifier_type": "open_alex", "identifier": "W1234567890"},
                {"identifier_type": "doi", "identifier": "10.1234/sample"},
            ],
            "enhancements": [
                {
                    "id": "52f3fc92-db0b-4e65-a18c-31d091242c3a",
                    "visibility": "public",
                    "source": "dummy",
                    "robot_version": "dummy",
                    "content": {
                        "enhancement_type": "bibliographic",
                        "authorship": [
                            {
                                "display_name": "Future E. Foundation",
                                "orcid": "0000-0001-2345-6789",
                                "position": "first",
                            }
                        ],
                        "cited_by_count": 3,
                        "created_date": "2016-08-23",
                        "publication_date": "2016-08-01",
                        "publication_year": 2016,
                        "publisher": "Research for Dummies",
                        "title": "Research!",
                    },
                },
                {
                    "id": "df4de714-269e-4d1c-abeb-16e260b029ec",
                    "visibility": "public",
                    "source": "dummy",
                    "robot_version": "dummy",
                    "content": {
                        "enhancement_type": "abstract",
                        "process": "other",
                        "abstract": "We did research.",
                    },
                },
                {
                    "id": "d4c2d0a0-af3a-4985-bcf6-1a9d58ab06ee",
                    "visibility": "public",
                    "source": "Toy Robot",
                    "robot_version": "0.1.0",
                    "content": {
                        "enhancement_type": "annotation",
                        "annotations": [
                            {
                                "annotation_type": "score",
                                "scheme": "meta:toy",
                                "label": "toy",
                                "score": 0.89,
                                "data": {"toy": "Bo Peep"},
                            }
                        ],
                    },
                },
            ],
        },
    }

    reference = destiny_sdk.references.Reference.from_es(es_reference)
    assert reference.id == uuid.UUID("7c54f72e-8833-484f-9000-4f403b13a243")
    assert reference.visibility == destiny_sdk.visibility.Visibility.PUBLIC
    assert len(reference.identifiers) == 2
    assert len(reference.enhancements) == 3
    assert reference.enhancements[0].reference_id == uuid.UUID(
        "7c54f72e-8833-484f-9000-4f403b13a243"
    )
    assert reference.enhancements[0].id == uuid.UUID(
        "52f3fc92-db0b-4e65-a18c-31d091242c3a"
    )



"""Tests for the EPPI parser."""

from pathlib import Path

from destiny_sdk.parsers.eppi_parser import EPPIParser


def test_parse_data():
    """Test that the parse_data method returns the expected output."""
    test_data_path = Path(__file__).parent.parent / "test_data"
    input_path = test_data_path / "eppi_report.json"
    output_path = test_data_path / "eppi_import.jsonl"

    import json

    parser = EPPIParser()
    with input_path.open() as f:
        data = json.load(f)
    references = parser.parse_data(data, robot_version="test-robot-version")

    with output_path.open() as f:
        expected_output = f.read()

    actual_output = "".join([ref.to_jsonl() + "\n" for ref in references])

    assert actual_output == expected_output


def test_parse_data_with_annotations():
    """Test that the parse_data method returns the output with annotations."""
    test_data_path = Path(__file__).parent.parent / "test_data"
    input_path = test_data_path / "eppi_report.json"
    output_path = test_data_path / "eppi_import_with_annotations.jsonl"

    import json

    parser = EPPIParser(tags=["test-tag", "another-tag"])
    with input_path.open() as f:
        data = json.load(f)
    references = parser.parse_data(data, robot_version="test-robot-version")

    with output_path.open() as f:
        expected_output = f.read()

    actual_output = "".join([ref.to_jsonl() + "\n" for ref in references])

    assert actual_output == expected_output



"""Send authenticated requests to Destiny Repository."""

import time
from collections.abc import Generator

import httpx
from pydantic import UUID4, HttpUrl

from destiny_sdk.robots import (
    BatchEnhancementRequestRead,
    BatchRobotResult,
    EnhancementRequestRead,
    RobotResult,
)

from .auth import create_signature


class HMACSigningAuth(httpx.Auth):
    """Client that adds an HMAC signature to a request."""

    requires_request_body = True

    def __init__(self, secret_key: str, client_id: UUID4) -> None:
        """
        Initialize the client.

        :param secret_key: the key to use when signing the request
        :type secret_key: str
        """
        self.secret_key = secret_key
        self.client_id = client_id

    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response]:
        """
        Add a signature to the given request.

        :param request: request to be sent with signature
        :type request: httpx.Request
        :yield: Generator for Request with signature headers set
        :rtype: Generator[httpx.Request, httpx.Response]
        """
        timestamp = time.time()
        signature = create_signature(
            self.secret_key, request.content, self.client_id, timestamp
        )
        request.headers["Authorization"] = f"Signature {signature}"
        request.headers["X-Client-Id"] = f"{self.client_id}"
        request.headers["X-Request-Timestamp"] = f"{timestamp}"
        yield request


class Client:
    """
    Client for interaction with the Destiny API.

    Current implementation only supports robot results.
    """

    def __init__(self, base_url: HttpUrl, secret_key: str, client_id: UUID4) -> None:
        """
        Initialize the client.

        :param base_url: The base URL for the Destiny Repository API.
        :type base_url: HttpUrl
        :param secret_key: The secret key for signing requests
        :type auth_method: str
        """
        self.session = httpx.Client(
            base_url=str(base_url).removesuffix("/").removesuffix("/v1") + "/v1",
            headers={"Content-Type": "application/json"},
            auth=HMACSigningAuth(secret_key=secret_key, client_id=client_id),
        )

    def send_robot_result(self, robot_result: RobotResult) -> EnhancementRequestRead:
        """
        Send a RobotResult to destiny repository.

        Signs the request with the client's secret key.

        :param robot_result: The Robot Result to send
        :type robot_result: RobotResult
        :return: The EnhancementRequestRead object from the response.
        :rtype: EnhancementRequestRead
        """
        response = self.session.post(
            f"/enhancement-requests/single-requests/{robot_result.request_id}/results/",
            json=robot_result.model_dump(mode="json"),
        )
        response.raise_for_status()
        return EnhancementRequestRead.model_validate(response.json())

    def send_batch_robot_result(
        self, batch_robot_result: BatchRobotResult
    ) -> BatchEnhancementRequestRead:
        """
        Send a BatchRobotResult to destiny repository.

        Signs the request with the client's secret key.

        :param batch_robot_result: The Batch Robot Result to send
        :type batch_robot_result: BatchRobotResult
        :return: The BatchEnhancementRequestRead object from the response.
        :rtype: BatchEnhancementRequestRead
        """
        response = self.session.post(
            f"/enhancement-requests/batch-requests/{batch_robot_result.request_id}/results/",
            json=batch_robot_result.model_dump(mode="json"),
        )
        response.raise_for_status()
        return BatchEnhancementRequestRead.model_validate(response.json())



"""Tests client authentication"""

import time
import uuid

import pytest
from destiny_sdk.client import Client, create_signature
from destiny_sdk.robots import EnhancementRequestRead, RobotError, RobotResult
from pytest_httpx import HTTPXMock


@pytest.fixture
def frozen_time(monkeypatch):
    def frozen_timestamp():
        return 12345453.32423

    monkeypatch.setattr(time, "time", frozen_timestamp)


def test_verify_hmac_headers_sent(httpx_mock: HTTPXMock, frozen_time) -> None:  # noqa: ARG001
    """Test that request is authorized with a signature."""
    fake_secret_key = "asdfhjgji94523q0uflsjf349wjilsfjd9q23"
    fake_robot_id = uuid.uuid4()
    fake_destiny_repository_url = "https://www.destiny-repository-lives-here.co.au/v1"

    fake_robot_result = RobotResult(
        request_id=uuid.uuid4(), error=RobotError(message="I can't fulfil this request")
    )

    expected_response_body = EnhancementRequestRead(
        reference_id=uuid.uuid4(),
        id=uuid.uuid4(),
        robot_id=uuid.uuid4(),
        request_status="failed",
    )

    expected_signature = create_signature(
        secret_key=fake_secret_key,
        request_body=fake_robot_result.model_dump_json().encode(),
        client_id=fake_robot_id,
        timestamp=time.time(),
    )

    httpx_mock.add_response(
        url=fake_destiny_repository_url
        + "/enhancement-requests/single-requests/"
        + f"{fake_robot_result.request_id}/results/",
        method="POST",
        match_headers={
            "Authorization": f"Signature {expected_signature}",
            "X-Client-Id": f"{fake_robot_id}",
            "X-Request-Timestamp": f"{time.time()}",
        },
        json=expected_response_body.model_dump(mode="json"),
    )

    Client(
        base_url=fake_destiny_repository_url,
        secret_key=fake_secret_key,
        client_id=fake_robot_id,
    ).send_robot_result(
        robot_result=fake_robot_result,
    )

    callback_request = httpx_mock.get_requests()
    assert len(callback_request) == 1



import destiny_sdk
import pytest
from pydantic import ValidationError


def test_valid_doi():
    obj = destiny_sdk.identifiers.DOIIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.DOI,
        identifier="10.1000/xyz123",
    )
    assert obj.identifier == "10.1000/xyz123"


def test_invalid_doi():
    with pytest.raises(ValidationError, match="String should match pattern"):
        destiny_sdk.identifiers.DOIIdentifier(
            identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.DOI,
            identifier="invalid_doi",
        )


def test_doi_url_removed():
    """Test that a DOI with a URL is fixed to just the DOI part."""
    obj = destiny_sdk.identifiers.DOIIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.DOI,
        identifier="http://doi.org/10.1000/xyz123",
    )
    assert obj.identifier == "10.1000/xyz123"


def test_valid_pmid():
    identifier = 123456

    obj = destiny_sdk.identifiers.PubMedIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.PM_ID,
        identifier=identifier,
    )
    assert obj.identifier == identifier


def test_invalid_pmid():
    with pytest.raises(ValidationError, match="Input should be a valid integer"):
        destiny_sdk.identifiers.PubMedIdentifier(
            identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.PM_ID,
            identifier="abc123",
        )


def test_valid_open_alex():
    valid_open_alex = "W123456789"
    obj = destiny_sdk.identifiers.OpenAlexIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OPEN_ALEX,
        identifier=valid_open_alex,
    )
    assert obj.identifier == valid_open_alex


def test_open_alex_url_removed():
    identitier = "W123456789"
    valid_openalex_with_url_https = f"https://openalex.org/{identitier}"

    obj = destiny_sdk.identifiers.OpenAlexIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OPEN_ALEX,
        identifier=valid_openalex_with_url_https,
    )

    assert obj.identifier == identitier

    valid_openalex_with_url_http = f"http://openalex.org/{identitier}"

    obj = destiny_sdk.identifiers.OpenAlexIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OPEN_ALEX,
        identifier=valid_openalex_with_url_http,
    )

    assert obj.identifier == identitier


def test_invalid_open_alex():
    with pytest.raises(ValidationError, match="String should match pattern"):
        destiny_sdk.identifiers.OpenAlexIdentifier(
            identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OPEN_ALEX,
            identifier="invalid-openalex",
        )


def test_valid_other_identifier():
    obj = destiny_sdk.identifiers.OtherIdentifier(
        identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OTHER,
        identifier="custom_identifier",
        other_identifier_name="custom_type",
    )
    assert obj.other_identifier_name == "custom_type"


def test_invalid_other_identifier_missing_name():
    with pytest.raises(
        ValidationError,
        match="Field required",
    ):
        destiny_sdk.identifiers.OtherIdentifier(
            identifier_type=destiny_sdk.identifiers.ExternalIdentifierType.OTHER,
            identifier="custom_identifier",
        )



"""Parser for a EPPI JSON export file."""

from typing import Any

from destiny_sdk.enhancements import (
    AbstractContentEnhancement,
    AbstractProcessType,
    AnnotationEnhancement,
    AnnotationType,
    AuthorPosition,
    Authorship,
    BibliographicMetadataEnhancement,
    BooleanAnnotation,
    EnhancementContent,
    EnhancementFileInput,
)
from destiny_sdk.identifiers import (
    DOIIdentifier,
    ExternalIdentifier,
    ExternalIdentifierType,
)
from destiny_sdk.references import ReferenceFileInput
from destiny_sdk.visibility import Visibility


class EPPIParser:
    """
    Parser for an EPPI JSON export file.

    See example here: https://eppi.ioe.ac.uk/cms/Portals/35/Maps/Examples/example_orignal.json
    """

    version = "1.0"

    def __init__(self, tags: list[str] | None = None) -> None:
        """
        Initialize the EPPIParser with optional tags.

        Args:
            tags (list[str] | None): Optional list of tags to annotate references.

        """
        self.tags = tags or []
        self.parser_source = f"destiny_sdk.eppi_parser@{self.version}"

    def _parse_identifiers(
        self, ref_to_import: dict[str, Any]
    ) -> list[ExternalIdentifier]:
        identifiers = []
        if doi := ref_to_import.get("DOI"):
            identifiers.append(
                DOIIdentifier(
                    identifier=doi,
                    identifier_type=ExternalIdentifierType.DOI,
                )
            )
        return identifiers

    def _parse_abstract_enhancement(
        self, ref_to_import: dict[str, Any]
    ) -> EnhancementContent | None:
        if abstract := ref_to_import.get("Abstract"):
            return AbstractContentEnhancement(
                process=AbstractProcessType.OTHER,
                abstract=abstract,
            )
        return None

    def _parse_bibliographic_enhancement(
        self, ref_to_import: dict[str, Any]
    ) -> EnhancementContent | None:
        title = ref_to_import.get("Title")
        publication_year = (
            int(year)
            if (year := ref_to_import.get("Year")) and year.isdigit()
            else None
        )
        publisher = ref_to_import.get("Publisher")
        authors_string = ref_to_import.get("Authors")

        authorships = []
        if authors_string:
            authors = [
                author.strip() for author in authors_string.split(";") if author.strip()
            ]
            for i, author_name in enumerate(authors):
                position = AuthorPosition.MIDDLE
                if i == 0:
                    position = AuthorPosition.FIRST
                if i == len(authors) - 1 and i > 0:
                    position = AuthorPosition.LAST

                authorships.append(
                    Authorship(
                        display_name=author_name,
                        position=position,
                    )
                )

        if not title and not publication_year and not publisher and not authorships:
            return None

        return BibliographicMetadataEnhancement(
            title=title,
            publication_year=publication_year,
            publisher=publisher,
            authorship=authorships if authorships else None,
        )

    def _create_annotation_enhancement(self) -> EnhancementContent | None:
        if not self.tags:
            return None
        annotations = [
            BooleanAnnotation(
                annotation_type=AnnotationType.BOOLEAN,
                scheme=self.parser_source,
                label=tag,
                value=True,
            )
            for tag in self.tags
        ]
        return AnnotationEnhancement(
            annotations=annotations,
        )

    def parse_data(
        self, data: dict, source: str | None = None, robot_version: str | None = None
    ) -> list[ReferenceFileInput]:
        """
        Parse an EPPI JSON export dict and return a list of ReferenceFileInput objects.

        Args:
            data (dict): Parsed EPPI JSON export data.
            source (str | None): Optional source string for deduplication/provenance.
            robot_version (str | None): Optional robot version string for provenance.
            Defaults to parser version.

        Returns:
            list[ReferenceFileInput]: List of parsed references from the data.

        """
        parser_source = source if source is not None else self.parser_source
        references = []
        for ref_to_import in data.get("References", []):
            enhancement_contents = [
                content
                for content in [
                    self._parse_abstract_enhancement(ref_to_import),
                    self._parse_bibliographic_enhancement(ref_to_import),
                    self._create_annotation_enhancement(),
                ]
                if content
            ]

            enhancements = [
                EnhancementFileInput(
                    source=parser_source,
                    visibility=Visibility.PUBLIC,
                    content=content,
                    enhancement_type=content.enhancement_type,
                    robot_version=robot_version,
                )
                for content in enhancement_contents
            ]

            references.append(
                ReferenceFileInput(
                    visibility=Visibility.PUBLIC,
                    identifiers=self._parse_identifiers(ref_to_import),
                    enhancements=enhancements,
                )
            )
        return references



"""Enhancement classes for the Destiny Repository."""

import datetime
from enum import StrEnum, auto
from typing import Annotated, Literal

from pydantic import UUID4, BaseModel, Field, HttpUrl

from destiny_sdk.core import _JsonlFileInputMixIn
from destiny_sdk.visibility import Visibility


class EnhancementType(StrEnum):
    """
    The type of enhancement.

    This is used to identify the type of enhancement in the `Enhancement` class.

    **Allowed values**:
    - `bibliographic`: Bibliographic metadata.
    - `abstract`: The abstract of a reference.
    - `annotation`: A free-form enhancement for tagging with labels.
    - `locations`: Locations where the reference can be found.
    - `full_text`: The full text of the reference. (To be implemeted)
    """

    BIBLIOGRAPHIC = auto()
    ABSTRACT = auto()
    ANNOTATION = auto()
    LOCATION = auto()
    FULL_TEXT = auto()


class AuthorPosition(StrEnum):
    """
    The position of an author in a list of authorships.

    Maps to the data from OpenAlex.

    **Allowed values**:
    - `first`: The first author.
    - `middle`: Any middle author
    - `last`: The last author
    """

    FIRST = auto()
    MIDDLE = auto()
    LAST = auto()


class Authorship(BaseModel):
    """
    Represents a single author and their association with a reference.

    This is a simplification of the OpenAlex [Authorship
    object](https://docs.openalex.org/api-entities/works/work-object/authorship-object)
    for our purposes.
    """

    display_name: str = Field(description="The display name of the author.")
    orcid: str | None = Field(default=None, description="The ORCid of the author.")
    position: AuthorPosition = Field(
        description="The position of the author within the list of authors."
    )


class BibliographicMetadataEnhancement(BaseModel):
    """
    An enhancement which is made up of bibliographic metadata.

    Generally this will be sourced from a database such as OpenAlex or similar.
    For directly contributed references, these may not be complete.
    """

    enhancement_type: Literal[EnhancementType.BIBLIOGRAPHIC] = (
        EnhancementType.BIBLIOGRAPHIC
    )
    authorship: list[Authorship] | None = Field(
        default=None,
        description="A list of `Authorships` belonging to this reference.",
    )
    cited_by_count: int | None = Field(
        default=None,
        description="""
(From OpenAlex) The number of citations to this work. These are the times that
other works have cited this work
""",
    )
    created_date: datetime.date | None = Field(
        default=None, description="The ISO8601 date this metadata record was created"
    )
    publication_date: datetime.date | None = Field(
        default=None, description="The date which the version of record was published."
    )
    publication_year: int | None = Field(
        default=None,
        description="The year in which the version of record was published.",
    )
    publisher: str | None = Field(
        default=None,
        description="The name of the entity which published the version of record.",
    )
    title: str | None = Field(default=None, description="The title of the reference.")


class AbstractProcessType(StrEnum):
    """
    The process used to acquire the abstract.

    **Allowed values**:
    - `uninverted`
    - `closed_api`
    - `other`
    """

    UNINVERTED = auto()
    CLOSED_API = auto()
    OTHER = auto()


class AbstractContentEnhancement(BaseModel):
    """
    An enhancement which is specific to the abstract of a reference.

    This is separate from the `BibliographicMetadata` for two reasons:

    1. Abstracts are increasingly missing from sources like OpenAlex, and may be
    backfilled from other sources, without the bibliographic metadata.
    2. They are also subject to copyright limitations in ways which metadata are
    not, and thus need separate visibility controls.
    """

    enhancement_type: Literal[EnhancementType.ABSTRACT] = EnhancementType.ABSTRACT
    process: AbstractProcessType = Field(
        description="The process used to acquire the abstract."
    )
    abstract: str = Field(description="The abstract of the reference.")


class AnnotationType(StrEnum):
    """
    The type of annotation.

    This is used to identify the type of annotation in the `Annotation` class.

    **Allowed values**:
    - `boolean`: An annotation which is the boolean application of a label across a
    reference.
    - `score`: An annotation which is a score for a label across a reference,
    without a boolean value.
    """

    BOOLEAN = auto()
    SCORE = auto()


class ScoreAnnotation(BaseModel):
    """
    An annotation which represents the score for a label.

    This is similar to a BooleanAnnotation, but lacks a boolean determination
    as to the application of the label.
    """

    annotation_type: Literal[AnnotationType.SCORE] = AnnotationType.SCORE
    scheme: str = Field(
        description="An identifier for the scheme of annotation",
        examples=["openalex:topic", "pubmed:mesh"],
    )
    label: str = Field(
        description="A high level label for this annotation like the name of the topic",
    )
    score: float = Field(description="""Score for this annotation""")
    data: dict = Field(
        default_factory=dict,
        description=(
            "An object representation of the annotation including any confidence scores"
            " or descriptions."
        ),
    )


class BooleanAnnotation(BaseModel):
    """
    An annotation is a way of tagging the content with a label of some kind.

    This class will probably be broken up in the future, but covers most of our
    initial cases.
    """

    annotation_type: Literal[AnnotationType.BOOLEAN] = AnnotationType.BOOLEAN
    scheme: str = Field(
        description="An identifier for the scheme of the annotation",
        examples=["openalex:topic", "pubmed:mesh"],
    )
    label: str = Field(
        description="A high level label for this annotation like the name of the topic",
    )
    value: bool = Field(description="""Boolean flag for this annotation""")
    score: float | None = Field(
        None, description="A confidence score for this annotation"
    )
    data: dict = Field(
        default_factory=dict,
        description="""
An object representation of the annotation including any confidence scores or
descriptions.
""",
    )


#: Union type for all annotations.
Annotation = Annotated[
    BooleanAnnotation | ScoreAnnotation, Field(discriminator="annotation_type")
]


class AnnotationEnhancement(BaseModel):
    """An enhancement which is composed of a list of Annotations."""

    enhancement_type: Literal[EnhancementType.ANNOTATION] = EnhancementType.ANNOTATION
    annotations: list[Annotation] = Field(min_length=1)


class DriverVersion(StrEnum):
    """
    The version based on the DRIVER guidelines versioning scheme.

    (Borrowed from OpenAlex)

    Allowed values:
    - `publishedVersion`: The document's version of record. This is the most
    authoritative version.
    - `acceptedVersion`: The document after having completed peer review and being
    officially accepted for publication. It will lack publisher formatting, but the
    content should be interchangeable with the that of the publishedVersion.
    - `submittedVersion`: the document as submitted to the publisher by the authors, but
    before peer-review. Its content may differ significantly from that of the accepted
    article.
    """

    PUBLISHED_VERSION = "publishedVersion"
    ACCEPTED_VERSION = "acceptedVersion"
    SUBMITTED_VERSION = "submittedVersion"
    OTHER = "other"


class Location(BaseModel):
    """
    A location where a reference can be found.

    This maps almost completely to the OpenAlex
    [Location object](https://docs.openalex.org/api-entities/works/work-object/location-object)
    """

    is_oa: bool | None = Field(
        default=None,
        description="""
(From OpenAlex): True if an Open Access (OA) version of this work is available
at this location. May be left as null if this is unknown (and thus)
treated effectively as `false`.
""",
    )
    version: DriverVersion | None = Field(
        default=None,
        description="""
The version (according to the DRIVER versioning scheme) of this location.
""",
    )
    landing_page_url: HttpUrl | None = Field(
        default=None,
        description="(From OpenAlex): The landing page URL for this location.",
    )
    pdf_url: HttpUrl | None = Field(
        default=None,
        description="""
(From OpenAlex): A URL where you can find this location as a PDF.
""",
    )
    license: str | None = Field(
        default=None,
        description="""
(From OpenAlex): The location's publishing license. This can be a Creative
Commons license such as cc0 or cc-by, a publisher-specific license, or null
which means we are not able to determine a license for this location.
""",
    )
    extra: dict | None = Field(
        default=None, description="Any extra metadata about this location"
    )


class LocationEnhancement(BaseModel):
    """
    An enhancement which describes locations where this reference can be found.

    This maps closely (almost exactly) to OpenAlex's locations.
    """

    enhancement_type: Literal[EnhancementType.LOCATION] = EnhancementType.LOCATION
    locations: list[Location] = Field(
        min_length=1,
        description="A list of locations where this reference can be found.",
    )


#: Union type for all enhancement content types.
EnhancementContent = Annotated[
    BibliographicMetadataEnhancement
    | AbstractContentEnhancement
    | AnnotationEnhancement
    | LocationEnhancement,
    Field(discriminator="enhancement_type"),
]


class Enhancement(_JsonlFileInputMixIn, BaseModel):
    """Core enhancement class."""

    id: UUID4 | None = Field(
        default=None,
        description=(
            "The ID of the enhancement. "
            "Populated by the repository when sending enhancements with references."
        ),
    )

    reference_id: UUID4 = Field(
        description="The ID of the reference this enhancement is associated with."
    )
    source: str = Field(
        description="The enhancement source for tracking provenance.",
    )
    visibility: Visibility = Field(
        description="The level of visibility of the enhancement"
    )
    robot_version: str | None = Field(
        default=None,
        description="The version of the robot that generated the content.",
    )
    derived_from: list[UUID4] | None = Field(
        default=None,
        description="List of enhancement IDs that this enhancement was derived from.",
    )
    content: Annotated[
        EnhancementContent,
        Field(
            discriminator="enhancement_type",
            description="The content of the enhancement.",
        ),
    ]


class EnhancementFileInput(BaseModel):
    """Enhancement model used to marshall a file input to new references."""

    source: str = Field(
        description="The enhancement source for tracking provenance.",
    )
    visibility: Visibility = Field(
        description="The level of visibility of the enhancement"
    )
    enhancement_type: EnhancementType = Field(description="The type of enhancement.")
    robot_version: str | None = Field(
        default=None,
        description="The version of the robot that generated the content.",
        # (Adam) Temporary alias for backwards compatibility for already prepared files
        # Next person who sees this can remove it :)
        alias="processor_version",
    )
    content: EnhancementContent = Field(
        discriminator="enhancement_type",
        description="The content of the enhancement.",
    )



"""The DESTINY SDK provides files for interacting with DESTINY repository."""

from . import (
    auth,
    client,
    enhancements,
    identifiers,
    imports,
    references,
    robots,
    visibility,
)

__all__ = [
    "auth",
    "client",
    "enhancements",
    "identifiers",
    "imports",
    "references",
    "robots",
    "visibility",
]



"""Visibility enum for repository data."""

from enum import StrEnum, auto


class Visibility(StrEnum):
    """
    The visibility of a data element in the repository.

    This is used to manage whether information should be publicly available or
    restricted (generally due to copyright constraints from publishers).

    TODO: Implement data governance layer to manage this.

    **Allowed values**:

    - `public`: Visible to the general public without authentication.
    - `restricted`: Requires authentication to be visible.
    - `hidden`: Is not visible, but may be passed to data mining processes.
    """

    PUBLIC = auto()
    RESTRICTED = auto()
    HIDDEN = auto()



"""HMAC authentication assistance methods."""

import hashlib
import hmac
import time
from typing import Protocol, Self
from uuid import UUID

from fastapi import HTTPException, Request, status
from pydantic import UUID4, BaseModel

FIVE_MINUTES = 60 * 5


class AuthException(HTTPException):
    """
    An exception related to HTTP authentication.

    Raised by implementations of the AuthMethod protocol.

    .. code-block:: python

        raise AuthException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Unable to parse authentication token.",
        )

    """


def create_signature(
    secret_key: str, request_body: bytes, client_id: UUID4, timestamp: float
) -> str:
    """
    Create an HMAC signature using SHA256.

    :param secret_key: secret key with which to encrypt message
    :type secret_key: bytes
    :param request_body: request body to be encrypted
    :type request_body: bytes
    :param client_id: client id to include in hmac
    :type: UUID4
    :param timestamp: timestamp for when the request is sent
    :type: float
    :return: encrypted hexdigest of the request body with the secret key
    :rtype: str
    """
    timestamp_bytes = f"{timestamp}".encode()
    signature_components = b":".join([request_body, client_id.bytes, timestamp_bytes])
    return hmac.new(
        secret_key.encode(), signature_components, hashlib.sha256
    ).hexdigest()


class HMACAuthorizationHeaders(BaseModel):
    """
    The HTTP authorization headers required for HMAC authentication.

    Expects the following headers to be present in the request

    - Authorization: Signature [request signature]
    - X-Client-Id: [UUID]
    - X-Request-Timestamp: [float]
    """

    signature: str
    client_id: UUID4
    timestamp: float

    @classmethod
    def from_request(cls, request: Request) -> Self:
        """
        Get the required headers for HMAC authentication.

        :param request: The incoming request
        :type request: Request
        :raises AuthException: Authorization header is missing
        :raises AuthException: Authorization type not supported
        :raises AuthException: X-Client-Id header is missing
        :raises AuthException: Client id format is invalid
        :raises AuthException: X-Request-Timestamp header is missing
        :raises AuthException: Request timestamp has expired
        :return: Header values necessary for authenticating the request
        :rtype: HMACAuthorizationHeaders
        """
        signature_header = request.headers.get("Authorization")

        if not signature_header:
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Authorization header missing.",
            )

        scheme, _, signature = signature_header.partition(" ")

        if scheme != "Signature":
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Authorization type not supported.",
            )

        client_id = request.headers.get("X-Client-Id")

        if not client_id:
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="X-Client-Id header missing",
            )

        try:
            UUID(client_id)
        except (ValueError, TypeError) as exc:
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid format for client id, expected UUID4.",
            ) from exc

        timestamp = request.headers.get("X-Request-Timestamp")

        if not timestamp:
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="X-Request-Timestamp header missing",
            )

        if (time.time() - float(timestamp)) > FIVE_MINUTES:
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Request timestamp has expired.",
            )

        return cls(signature=signature, client_id=client_id, timestamp=timestamp)


class HMACAuthMethod(Protocol):
    """
    Protocol for HMAC auth methods, enforcing the implmentation of __call__().

    This allows FastAPI to call class instances as depenedencies in FastAPI routes,
    see https://fastapi.tiangolo.com/advanced/advanced-dependencies

        .. code-block:: python

            auth = HMACAuthMethod()

            router = APIRouter(
                prefix="/robots", tags=["robot"], dependencies=[Depends(auth)]
            )
    """

    async def __call__(self, request: Request) -> bool:
        """
        Callable interface to allow use as a dependency.

        :param request: The request to verify
        :type request: Request
        :raises NotImplementedError: __call__() method has not been implemented.
        :return: True if authorization is successful.
        :rtype: bool
        """
        raise NotImplementedError


class HMACAuth(HMACAuthMethod):
    """Adds HMAC auth when used as a router or endpoint dependency."""

    def __init__(self, secret_key: str) -> None:
        """Initialise HMAC auth with a given secret key."""
        self.secret_key = secret_key

    async def __call__(self, request: Request) -> bool:
        """Perform Authorization check."""
        auth_headers = HMACAuthorizationHeaders.from_request(request)
        request_body = await request.body()
        expected_signature = create_signature(
            self.secret_key,
            request_body,
            auth_headers.client_id,
            auth_headers.timestamp,
        )

        if not hmac.compare_digest(auth_headers.signature, expected_signature):
            raise AuthException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Signature is invalid."
            )

        return True


class BypassHMACAuth(HMACAuthMethod):
    """
    A fake auth class that will always respond successfully.

    Intended for use in local environments and for testing.

    Not for production use!
    """

    async def __call__(
        self,
        request: Request,  # noqa: ARG002
    ) -> bool:
        """Bypass Authorization check."""
        return True
